@@ -2,14 +2,16 @@
 /*
  * Copyright (C) 2019, Google Inc.
  *
- * request.h - Request class private data
+ * Request class private data
  */
-#ifndef __LIBCAMERA_INTERNAL_REQUEST_H__
-#define __LIBCAMERA_INTERNAL_REQUEST_H__
+
+#pragma once
 
 #include <chrono>
 #include <map>
 #include <memory>
+#include <stdint.h>
+#include <unordered_set>
 
 #include <libcamera/base/event_notifier.h>
 #include <libcamera/base/timer.h>
@@ -62,5 +64,3 @@ private:
 };
 
 } /* namespace libcamera */
-
-#endif /* __LIBCAMERA_INTERNAL_REQUEST_H__ */
