@@ -2,17 +2,22 @@
 /*
  * Copyright (C) 2019, Google Inc.
  *
- * uvcvideo.cpp - Pipeline handler for uvcvideo devices
+ * Pipeline handler for uvcvideo devices
  */
 
 #include <algorithm>
+#include <bitset>
+#include <cmath>
 #include <fstream>
-#include <iomanip>
-#include <math.h>
+#include <map>
 #include <memory>
-#include <tuple>
+#include <optional>
+#include <set>
+#include <string>
+#include <vector>
 
 #include <libcamera/base/log.h>
+#include <libcamera/base/mutex.h>
 #include <libcamera/base/utils.h>
 
 #include <libcamera/camera.h>
@@ -44,14 +49,18 @@ public:
 	int init(MediaDevice *media);
 	void addControl(uint32_t cid, const ControlInfo &v4l2info,
 			ControlInfoMap::Map *ctrls);
-	void bufferReady(FrameBuffer *buffer);
+	void imageBufferReady(FrameBuffer *buffer);
 
 	const std::string &id() const { return id_; }
 
+	Mutex openLock_;
 	std::unique_ptr<V4L2VideoDevice> video_;
 	Stream stream_;
 	std::map<PixelFormat, std::vector<SizeRange>> formats_;
 
+	std::optional<v4l2_exposure_auto_type> autoExposureMode_;
+	std::optional<v4l2_exposure_auto_type> manualExposureMode_;
+
 private:
 	bool generateId();
 
@@ -89,16 +98,39 @@ public:
 	bool match(DeviceEnumerator *enumerator) override;
 
 private:
-	int processControl(ControlList *controls, unsigned int id,
-			   const ControlValue &value);
+	int processControl(const UVCCameraData *data, ControlList *controls,
+			   unsigned int id, const ControlValue &value);
 	int processControls(UVCCameraData *data, Request *request);
 
+	bool acquireDevice(Camera *camera) override;
+	void releaseDevice(Camera *camera) override;
+
 	UVCCameraData *cameraData(Camera *camera)
 	{
 		return static_cast<UVCCameraData *>(camera->_d());
 	}
 };
 
+namespace {
+
+std::optional<controls::ExposureTimeModeEnum> v4l2ToExposureMode(int32_t x)
+{
+	using namespace controls;
+
+	switch (x) {
+	case V4L2_EXPOSURE_AUTO:
+	case V4L2_EXPOSURE_APERTURE_PRIORITY:
+		return ExposureTimeModeAuto;
+	case V4L2_EXPOSURE_MANUAL:
+	case V4L2_EXPOSURE_SHUTTER_PRIORITY:
+		return ExposureTimeModeManual;
+	default:
+		return {};
+	}
+}
+
+} /* namespace */
+
 UVCCameraConfiguration::UVCCameraConfiguration(UVCCameraData *data)
 	: CameraConfiguration(), data_(data)
 {
@@ -158,9 +190,29 @@ CameraConfiguration::Status UVCCameraConfiguration::validate()
 	format.fourcc = data_->video_->toV4L2PixelFormat(cfg.pixelFormat);
 	format.size = cfg.size;
 
-	int ret = data_->video_->tryFormat(&format);
-	if (ret)
-		return Invalid;
+	/*
+	 * For power-consumption reasons video_ is closed when the camera is not
+	 * acquired. Open it here if necessary.
+	 */
+	{
+		bool opened = false;
+
+		MutexLocker locker(data_->openLock_);
+
+		if (!data_->video_->isOpen()) {
+			int ret = data_->video_->open();
+			if (ret)
+				return Invalid;
+
+			opened = true;
+		}
+
+		int ret = data_->video_->tryFormat(&format);
+		if (opened)
+			data_->video_->close();
+		if (ret)
+			return Invalid;
+	}
 
 	cfg.stride = format.planes[0].bpl;
 	cfg.frameSize = format.planes[0].size;
@@ -260,8 +312,8 @@ void PipelineHandlerUVC::stopDevice(Camera *camera)
 	data->video_->releaseBuffers();
 }
 
-int PipelineHandlerUVC::processControl(ControlList *controls, unsigned int id,
-				       const ControlValue &value)
+int PipelineHandlerUVC::processControl(const UVCCameraData *data, ControlList *controls,
+				       unsigned int id, const ControlValue &value)
 {
 	uint32_t cid;
 
@@ -271,12 +323,14 @@ int PipelineHandlerUVC::processControl(ControlList *controls, unsigned int id,
 		cid = V4L2_CID_CONTRAST;
 	else if (id == controls::Saturation)
 		cid = V4L2_CID_SATURATION;
-	else if (id == controls::AeEnable)
+	else if (id == controls::ExposureTimeMode)
 		cid = V4L2_CID_EXPOSURE_AUTO;
 	else if (id == controls::ExposureTime)
 		cid = V4L2_CID_EXPOSURE_ABSOLUTE;
 	else if (id == controls::AnalogueGain)
 		cid = V4L2_CID_GAIN;
+	else if (id == controls::Gamma)
+		cid = V4L2_CID_GAMMA;
 	else
 		return -EINVAL;
 
@@ -293,22 +347,33 @@ int PipelineHandlerUVC::processControl(ControlList *controls, unsigned int id,
 	case V4L2_CID_BRIGHTNESS: {
 		float scale = std::max(max - def, def - min);
 		float fvalue = value.get<float>() * scale + def;
-		controls->set(cid, static_cast<int32_t>(lroundf(fvalue)));
+		controls->set(cid, static_cast<int32_t>(std::lround(fvalue)));
 		break;
 	}
 
 	case V4L2_CID_SATURATION: {
 		float scale = def - min;
 		float fvalue = value.get<float>() * scale + min;
-		controls->set(cid, static_cast<int32_t>(lroundf(fvalue)));
+		controls->set(cid, static_cast<int32_t>(std::lround(fvalue)));
 		break;
 	}
 
 	case V4L2_CID_EXPOSURE_AUTO: {
-		int32_t ivalue = value.get<bool>()
-			       ? V4L2_EXPOSURE_APERTURE_PRIORITY
-			       : V4L2_EXPOSURE_MANUAL;
-		controls->set(V4L2_CID_EXPOSURE_AUTO, ivalue);
+		std::optional<v4l2_exposure_auto_type> mode;
+
+		switch (value.get<int32_t>()) {
+		case controls::ExposureTimeModeAuto:
+			mode = data->autoExposureMode_;
+			break;
+		case controls::ExposureTimeModeManual:
+			mode = data->manualExposureMode_;
+			break;
+		}
+
+		if (!mode)
+			return -EINVAL;
+
+		controls->set(V4L2_CID_EXPOSURE_AUTO, static_cast<int32_t>(*mode));
 		break;
 	}
 
@@ -327,10 +392,14 @@ int PipelineHandlerUVC::processControl(ControlList *controls, unsigned int id,
 		}
 
 		float fvalue = (value.get<float>() - p) / m;
-		controls->set(cid, static_cast<int32_t>(lroundf(fvalue)));
+		controls->set(cid, static_cast<int32_t>(std::lround(fvalue)));
 		break;
 	}
 
+	case V4L2_CID_GAMMA:
+		controls->set(cid, static_cast<int32_t>(std::lround(value.get<float>() * 100)));
+		break;
+
 	default: {
 		int32_t ivalue = value.get<int32_t>();
 		controls->set(cid, ivalue);
@@ -346,7 +415,7 @@ int PipelineHandlerUVC::processControls(UVCCameraData *data, Request *request)
 	ControlList controls(data->video_->controls());
 
 	for (const auto &[id, value] : request->controls())
-		processControl(&controls, id, value);
+		processControl(data, &controls, id, value);
 
 	for (const auto &ctrl : controls)
 		LOG(UVC, Debug)
@@ -411,6 +480,23 @@ bool PipelineHandlerUVC::match(DeviceEnumerator *enumerator)
 	return true;
 }
 
+bool PipelineHandlerUVC::acquireDevice(Camera *camera)
+{
+	UVCCameraData *data = cameraData(camera);
+
+	MutexLocker locker(data->openLock_);
+
+	return data->video_->open() == 0;
+}
+
+void PipelineHandlerUVC::releaseDevice(Camera *camera)
+{
+	UVCCameraData *data = cameraData(camera);
+
+	MutexLocker locker(data->openLock_);
+	data->video_->close();
+}
+
 int UVCCameraData::init(MediaDevice *media)
 {
 	int ret;
@@ -432,7 +518,7 @@ int UVCCameraData::init(MediaDevice *media)
 	if (ret)
 		return ret;
 
-	video_->bufferReady.connect(this, &UVCCameraData::bufferReady);
+	video_->bufferReady.connect(this, &UVCCameraData::imageBufferReady);
 
 	/* Generate the camera ID. */
 	if (!generateId()) {
@@ -510,8 +596,19 @@ int UVCCameraData::init(MediaDevice *media)
 		addControl(cid, info, &ctrls);
 	}
 
+	if (autoExposureMode_ && manualExposureMode_) {
+		/* \todo Move this to the Camera class */
+		ctrls[&controls::AeEnable] = ControlInfo(false, true, true);
+	}
+
 	controlInfo_ = ControlInfoMap(std::move(ctrls), controls::controls);
 
+	/*
+	 * Close to allow camera to go into runtime-suspend, video_ will be
+	 * re-opened from acquireDevice() and validate().
+	 */
+	video_->close();
+
 	return 0;
 }
 
@@ -597,7 +694,7 @@ void UVCCameraData::addControl(uint32_t cid, const ControlInfo &v4l2Info,
 		id = &controls::Saturation;
 		break;
 	case V4L2_CID_EXPOSURE_AUTO:
-		id = &controls::AeEnable;
+		id = &controls::ExposureTimeMode;
 		break;
 	case V4L2_CID_EXPOSURE_ABSOLUTE:
 		id = &controls::ExposureTime;
@@ -605,11 +702,15 @@ void UVCCameraData::addControl(uint32_t cid, const ControlInfo &v4l2Info,
 	case V4L2_CID_GAIN:
 		id = &controls::AnalogueGain;
 		break;
+	case V4L2_CID_GAMMA:
+		id = &controls::Gamma;
+		break;
 	default:
 		return;
 	}
 
 	/* Map the control info. */
+	const std::vector<ControlValue> &v4l2Values = v4l2Info.values();
 	int32_t min = v4l2Info.min().get<int32_t>();
 	int32_t max = v4l2Info.max().get<int32_t>();
 	int32_t def = v4l2Info.def().get<int32_t>();
@@ -647,10 +748,79 @@ void UVCCameraData::addControl(uint32_t cid, const ControlInfo &v4l2Info,
 		};
 		break;
 
-	case V4L2_CID_EXPOSURE_AUTO:
-		info = ControlInfo{ false, true, true };
-		break;
+	case V4L2_CID_EXPOSURE_AUTO: {
+		/*
+		 * From the V4L2_CID_EXPOSURE_AUTO documentation:
+		 *
+		 * ------------------------------------------------------------
+		 * V4L2_EXPOSURE_AUTO:
+		 * Automatic exposure time, automatic iris aperture.
+		 *
+		 * V4L2_EXPOSURE_MANUAL:
+		 * Manual exposure time, manual iris.
+		 *
+		 * V4L2_EXPOSURE_SHUTTER_PRIORITY:
+		 * Manual exposure time, auto iris.
+		 *
+		 * V4L2_EXPOSURE_APERTURE_PRIORITY:
+		 * Auto exposure time, manual iris.
+		 *-------------------------------------------------------------
+		 *
+		 * ExposureTimeModeAuto = { V4L2_EXPOSURE_AUTO,
+		 * 			    V4L2_EXPOSURE_APERTURE_PRIORITY }
+		 *
+		 *
+		 * ExposureTimeModeManual = { V4L2_EXPOSURE_MANUAL,
+		 *			      V4L2_EXPOSURE_SHUTTER_PRIORITY }
+		 */
 
+		std::bitset<
+			std::max(V4L2_EXPOSURE_AUTO,
+			std::max(V4L2_EXPOSURE_APERTURE_PRIORITY,
+			std::max(V4L2_EXPOSURE_MANUAL,
+				 V4L2_EXPOSURE_SHUTTER_PRIORITY))) + 1
+		> exposureModes;
+		std::optional<controls::ExposureTimeModeEnum> lcDef;
+
+		for (const ControlValue &value : v4l2Values) {
+			const auto x = value.get<int32_t>();
+
+			if (0 <= x && static_cast<std::size_t>(x) < exposureModes.size()) {
+				exposureModes[x] = true;
+
+				if (x == def)
+					lcDef = v4l2ToExposureMode(x);
+			}
+		}
+
+		if (exposureModes[V4L2_EXPOSURE_AUTO])
+			autoExposureMode_ = V4L2_EXPOSURE_AUTO;
+		else if (exposureModes[V4L2_EXPOSURE_APERTURE_PRIORITY])
+			autoExposureMode_ = V4L2_EXPOSURE_APERTURE_PRIORITY;
+
+		if (exposureModes[V4L2_EXPOSURE_SHUTTER_PRIORITY])
+			manualExposureMode_ = V4L2_EXPOSURE_SHUTTER_PRIORITY;
+		else if (exposureModes[V4L2_EXPOSURE_MANUAL])
+			manualExposureMode_ = V4L2_EXPOSURE_MANUAL;
+
+		std::array<ControlValue, 2> values;
+		std::size_t count = 0;
+
+		if (autoExposureMode_)
+			values[count++] = controls::ExposureTimeModeAuto;
+
+		if (manualExposureMode_)
+			values[count++] = controls::ExposureTimeModeManual;
+
+		if (count == 0)
+			return;
+
+		info = ControlInfo{
+			Span<const ControlValue>{ values.data(), count },
+			!lcDef ? values.front() : *lcDef,
+		};
+		break;
+	}
 	case V4L2_CID_EXPOSURE_ABSOLUTE:
 		/*
 		 * ExposureTime is in units of 1 µs, and UVC expects
@@ -689,6 +859,15 @@ void UVCCameraData::addControl(uint32_t cid, const ControlInfo &v4l2Info,
 		break;
 	}
 
+	case V4L2_CID_GAMMA:
+		/* UVC gamma is in units of 1/100 gamma. */
+		info = ControlInfo{
+			{ min / 100.0f },
+			{ max / 100.0f },
+			{ def / 100.0f }
+		};
+		break;
+
 	default:
 		info = v4l2Info;
 		break;
@@ -697,7 +876,7 @@ void UVCCameraData::addControl(uint32_t cid, const ControlInfo &v4l2Info,
 	ctrls->emplace(id, info);
 }
 
-void UVCCameraData::bufferReady(FrameBuffer *buffer)
+void UVCCameraData::imageBufferReady(FrameBuffer *buffer)
 {
 	Request *request = buffer->request();
 
@@ -709,6 +888,6 @@ void UVCCameraData::bufferReady(FrameBuffer *buffer)
 	pipe()->completeRequest(request);
 }
 
-REGISTER_PIPELINE_HANDLER(PipelineHandlerUVC)
+REGISTER_PIPELINE_HANDLER(PipelineHandlerUVC, "uvcvideo")
 
 } /* namespace libcamera */
