@@ -2,7 +2,7 @@
 /*
  * Copyright (C) 2019, Google Inc.
  *
- * process.cpp - Process object
+ * Process object
  */
 
 #include "libcamera/internal/process.h"
@@ -10,7 +10,6 @@
 #include <algorithm>
 #include <dirent.h>
 #include <fcntl.h>
-#include <iostream>
 #include <list>
 #include <signal.h>
 #include <string.h>
@@ -76,7 +75,7 @@ void ProcessManager::sighandler()
 		return;
 	}
 
-	for (auto it = processes_.begin(); it != processes_.end(); ) {
+	for (auto it = processes_.begin(); it != processes_.end();) {
 		Process *process = *it;
 
 		int wstatus;
@@ -189,7 +188,6 @@ const struct sigaction &ProcessManager::oldsa() const
 	return oldsa_;
 }
 
-
 /**
  * \class Process
  * \brief Process object
@@ -261,7 +259,21 @@ int Process::start(const std::string &path,
 		if (isolate())
 			_exit(EXIT_FAILURE);
 
-		closeAllFdsExcept(fds);
+		std::vector<int> v(fds);
+		v.push_back(STDERR_FILENO);
+		closeAllFdsExcept(v);
+
+		const auto tryDevNullLowestFd = [](int expected, int oflag) {
+			int fd = open("/dev/null", oflag);
+			if (fd < 0)
+				_exit(EXIT_FAILURE);
+			if (fd != expected)
+				close(fd);
+		};
+
+		tryDevNullLowestFd(STDIN_FILENO, O_RDONLY);
+		tryDevNullLowestFd(STDOUT_FILENO, O_WRONLY);
+		tryDevNullLowestFd(STDERR_FILENO, O_WRONLY);
 
 		const char *file = utils::secure_getenv("LIBCAMERA_LOG_FILE");
 		if (file && strcmp(file, "syslog"))
@@ -271,12 +283,12 @@ int Process::start(const std::string &path,
 		unsigned int len = args.size();
 		argv[0] = path.c_str();
 		for (unsigned int i = 0; i < len; i++)
-			argv[i+1] = args[i].c_str();
-		argv[len+1] = nullptr;
+			argv[i + 1] = args[i].c_str();
+		argv[len + 1] = nullptr;
 
 		execv(path.c_str(), (char **)argv);
 
-		exit(EXIT_FAILURE);
+		_exit(EXIT_FAILURE);
 	}
 }
 
